#include "sqlite3_wrapper.h"
#include "str_fields_check.h"
#include <engine_engine_number_9/baza_throw.h>
#include <assert.h>
#include <limits.h>

SqliteConnection::SqliteConnection(const std::string &file_path) {
    int ret = sqlite3_open(file_path.c_str(), &hand);
    if (ret != 0) {
        een9_THROW(std::string("Can't open sqlite3 database ") + sqlite3_errstr(ret));
    }
}

SqliteConnection::~SqliteConnection() {
    if (sqlite3_close(hand) != 0) {abort();}
}

void sqlite_single_statement(sqlite3* db_hand, const std::string& req_statement,
                             const std::vector<std::pair<int, int64_t>>& int64_binds,
                             const std::vector<std::pair<int, std::string>>& text8_binds) {
    sqlite3_stmt* stmt_obj = NULL;
    int ret = sqlite3_prepare_v2(db_hand, req_statement.c_str(), -1, &stmt_obj, NULL);
    if (ret != 0) {
        int err_pos = sqlite3_error_offset(db_hand);
        een9_THROW("Compilation of request\n" + req_statement + "\nfailed" +
            ((err_pos >= 0) ? " with offset " + std::to_string(err_pos) : ""));
    }
    een9_ASSERT(ret == 0, "Can't compile request expression");
    assert(sqlite3_errcode(db_hand) == SQLITE_OK);
    struct Guard1{sqlite3_stmt*& r; ~Guard1(){if (sqlite3_finalize(r) != 0) {abort();}}} guard1{stmt_obj};
    for (const std::pair<int, int64_t>& bv: int64_binds) {
        ret = sqlite3_bind_int64(stmt_obj, bv.first, bv.second);
        een9_ASSERT(ret, "Can't bind to parameter #" + std::to_string(bv.first));
    }
    for (const std::pair<int, std::string>& bv: text8_binds) {
        een9_ASSERT(is_orthodox_string(bv.second), "Can't bind this string to parameter");
        een9_ASSERT(bv.second.size() + 1 < INT_MAX, "Ah, oh, senpai, your string is toooo huge");
        ret = sqlite3_bind_text(stmt_obj, bv.first, bv.second.c_str(), (int)bv.second.size(), SQLITE_STATIC);
    }
    while (true) {
        ret = sqlite3_step(stmt_obj);
        if (ret == SQLITE_DONE)
            break;
        if (ret != SQLITE_ROW) {
            printf("sqlite_row error!!!\n");
            printf("%s\n", sqlite3_errmsg(db_hand));
            break;
        }
        int cc = sqlite3_column_count(stmt_obj);
        std::vector<int> types(cc);
        for (int i = 0; i < cc; i++) {
            types[i] = sqlite3_column_type(stmt_obj, i);
        }
        printf("Column: |");
        for (int i = 0; i < cc; i++) {
            switch (types[i]) {
#define ccase(tname) case SQLITE_ ## tname: printf(" " #tname " |"); break;
                ccase(INTEGER)
                ccase(FLOAT)
                ccase(BLOB)
                ccase(NULL)
                case SQLITE3_TEXT:
                    printf(" TEXT |"); break;
                default:
                    een9_THROW("AAAAAA");
            }
        }
        printf("\n");
        printf("Values: | ");
        for (int i = 0; i < cc; i++) {
            if (types[i] == SQLITE_INTEGER) {
                printf("%lld | ", sqlite3_column_int64(stmt_obj, i));
            } else if (types[i] == SQLITE_FLOAT) {
                printf("%lf | ", sqlite3_column_double(stmt_obj, i));
            } else if (types[i] == SQLITE_BLOB) {
                const void* blob = sqlite3_column_blob(stmt_obj, i);
                een9_ASSERT(sqlite3_errcode(db_hand) == SQLITE_OK, "oom in sqlite3_column_blob");
                size_t sz = sqlite3_column_bytes(stmt_obj, i);
                printf("Blob of size %lu | ", sz);
            } else if (types[i] == SQLITE_NULL) {
                printf("NULL | ");
            } else {
                const unsigned char* text = sqlite3_column_text(stmt_obj, i);
                een9_ASSERT(text, "oom in sqlite3_column_text");
                printf("%s | ", (const char*)text);
                // todo: THIS F. B.S. IS NOT SAFE
            }
        }
        printf("\n");
    }
    printf("Request steps are done\n");
}
